/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Calendar Palm Sync code.
 *
 * The Initial Developer of the Original Code is
 * Michiel van Leeuwen (mvl@exedo.nl).
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function nsSyncSession() { }

nsSyncSession.prototype = {

  get fastSync()          { return this.mFastSync; },

  init: function (aSyncClient) {
    this.mClient = aSyncClient;

    this.mCommandList = new Array();

    dump(this.mClient.identifier+"\n");
  },

  /**
   * The merge to local logic:
   *  - Loop over all remote record
   *    - Find local record with the right sync id
   *      - not found -> create
   *      - found, but different -> update local
   *  - Loop over local records with a sync id
   *    - check if remote record with that id exist
   *      - not found -> delete local
   */
  
  /**
   * This should be async. By the time we do mergeToRemote, we don't
   * want the local list to have changed. Further, we don't want the
   * local list change while iterating it
   */

  fastMergeToLocal: function () {
dump("fastMergeToLocal\n");
    var c;
    
    var remoteRecords = this.mClient.remoteRecords;
    while (remoteRecords.hasMoreElements()) {
      var remoteRecord = remoteRecords.getNext().QueryInterface(Components.interfaces.nsISyncRecord);

      // Get the matching local record
      var localRecord = this.mClient.getLocalRecordBySyncId(remoteRecord.id);

      if (localRecord &&
          !remoteRecord.deleted &&
          !localRecord.dirty &&
          remoteRecord.dirty) {
        c = new Object();
        c.command = "UpdateLocal";
        c.param1 = remoteRecord;
        c.param2 = localRecord;
        this.mCommandList.push(c);
      }

      else if (localRecord &&
               !remoteRecord.deleted &&
               localRecord.dirty &&
               remoteRecord.dirty &&
               !this.mClient.compare(localRecord, remoteRecord)) {
        // There is a conflict.... The changed localrecord doesn't match
        // the remote record.
        // XXX id's
        c = new Object();
        c.command = "AddLocal";
        c.param1 = remoteRecord;
        this.mCommandList.push(c);
      }

      else if (localRecord &&
               remoteRecord.deleted) {
        c = new Object();
        c.command = "RemoveLocal";
        c.param1 = localRecord;
        this.mCommandList.push(c);
      }
      else if (!localRecord &&
               !remoteRecord.deleted) {
        c = new Object();
        c.command = "AddLocal";
        c.param1 = remoteRecord;
        this.mCommandList.push(c);
      }
    }
  },

  slowMergeToLocal: function () {
dump("slowMergeToLocal\n");
    var c;

    // This is slow sync, we can't trust the dirty record on the remote.
    var remoteRecords = this.mClient.remoteRecords;
    while (remoteRecords.hasMoreElements()) {
      var remoteRecord = remoteRecords.getNext().QueryInterface(Components.interfaces.nsISyncRecord);
dump("rid: "+remoteRecord.id+"\n");
      var localRecord = this.mClient.getLocalRecordBySyncId(remoteRecord.id);

      if (localRecord &&
          !remoteRecord.deleted &&
          !localRecord.dirty &&
          !this.mClient.compare(localRecord, remoteRecord)) {
        c = new Object();
        c.command = "UpdateLocal";
        c.param1 = remoteRecord;
        c.param2 = localRecord;
        this.mCommandList.push(c);
      }
      else if (localRecord &&
               !remoteRecord.deleted &&
               localRecord.dirty &&
               !this.mClient.compare(localRecord, remoteRecord)) {
        // There is a conflict.... The changed localrecord doesn't match
        // the remote record.
        // XXX id's
        c = new Object();
        c.command = "AddLocal";
        c.param1 = remoteRecord;
        this.mCommandList.push(c);
      }

      else if (localRecord &&
               remoteRecord.deleted) {
        c = new Object();
        c.command = "RemoveLocal";
        c.param1 = localRecord;
        this.mCommandList.push(c);
      }

      else if (!localRecord &&
               !remoteRecord.deleted) {
        c = new Object();
        c.command = "AddLocal";
        c.param1 = remoteRecord;
        this.mCommandList.push(c);
      }
    }

/*
    var localRecords = this.mClient.localRecords;
    while (localRecords.hasMoreElements()) {
      localRecord = localRecords.getNext().QueryInterface(Components.interfaces.nsISyncRecord);
      if (localRecord.id) {
        remoteRecord = this.mClient.getRemoteRecordBySyncId(localRecord.id);

        if (!remoteRecord)
          this.mClient.removeLocal(localRecord);
      }
    }
*/
  },

  fastMergeToRemote: function () { 
dump("fastMergeToRemote\n");
    var c;

    var localRecords = this.mClient.localRecords;
    while (localRecords.hasMoreElements()) {
      var localRecord = localRecords.getNext().QueryInterface(Components.interfaces.nsISyncRecord);

      // Get the matching remote record
      var remoteRecord = this.mClient.getRemoteRecordBySyncId(localRecord.id);

      if (remoteRecord &&
          !localRecord.deleted &&
          !remoteRecord.dirty &&
          localRecord.dirty) {
        c = new Object();
        c.command = "UpdateRemote";
        c.param1 = localRecord;
        c.param2 = remoteRecord;
        this.mCommandList.push(c);
      }

      else if (remoteRecord &&
               !localRecord.deleted &&
               remoteRecord.dirty &&
               localRecord.dirty &&
               !this.mClient.compare(localRecord, remoteRecord)) {
        // There is a conflict.... The changed localrecord doesn't match
        // the remote record.
        // XXX id's
        c = new Object();
        c.command = "AddRemote";
        c.param1 = localRecord;
        this.mCommandList.push(c);
      }

      else if (remoteRecord &&
               localRecord.deleted) {
        c = new Object();
        c.command = "RemoveRemote";
        c.param1 = remoteRecord;
        this.mCommandList.push(c);
      }

      else if (!remoteRecord &&
               !localRecord.deleted) {
        c = new Object();
        c.command = "AddRemote";
        c.param1 = localRecord;
        this.mCommandList.push(c);
      }
    }
  },

  slowMergeToRemote: function () { 
dump("slowMergeToRemote\n");
    var c;

    // In case of slow merge to remote, the dirty bits of the remote records
    // can't be trusted. So we never know is a record has changed. We can't
    // overwrite them!

    var localRecords = this.mClient.localRecords;
    while (localRecords.hasMoreElements()) {
      var localRecord = localRecords.getNext().QueryInterface(Components.interfaces.nsISyncRecord);
      var remoteRecord = this.mClient.getRemoteRecordBySyncId(localRecord.id);

      if (remoteRecord &&
          !localRecord.deleted &&
          localRecord.dirty &&
          !this.mClient.compare(localRecord, remoteRecord)) {
        // There is a conflict.... The changed localrecord doesn't match
        // the remote record.
        // XXX id's
        c = new Object();
        c.command = "AddRemote";
        c.param1 = localRecord;
        this.mCommandList.push(c);
      }

      else if (remoteRecord &&
               localRecord.deleted) {
        c = new Object();
        c.command = "RemoveRemote";
        c.param1 = remoteRecord;
        this.mCommandList.push(c);
      }

      else if (!remoteRecord &&
               !localRecord.deleted) {
        c = new Object();
        c.command = "AddRemote";
        c.param1 = localRecord;
        this.mCommandList.push(c);
      }
    }

/*    var remoteRecords = this.mClient.remoteRecords;
    while (remoteRecords.hasMoreElements()) {
      remoteRecord = remoteRecords.getNext().QueryInterface(Components.interfaces.nsISyncRecord);
      if (remoteRecord.id) {
        localRecord = this.mClient.getLocalRecordBySyncId(remoteRecord.id);

        if (!remoteRecord)
          this.mClient.removeRemote(remoteRecord);
      }
    }
*/
  },

  processList: function () {
    var c;

    while ((c = this.mCommandList.shift())) {
      dump("c: "+c.command+"\n");
      switch (c.command) {
      case "AddLocal":
        this.mClient.addLocal(c.param1);
        break;
      case "UpdateLocal":
        this.mClient.updateLocal(c.param1, c.param2);
        break;
      case "RemoveLocal":
        this.mClient.removeLocal(c.param1);
        break;
      case "AddRemote":
        this.mClient.addRemote(c.param1);
        break;
      case "UpdateRemote":
        this.mClient.updateRemote(c.param1, c.param2);
        break;
      case "RemoveRemote":
        this.mClient.removeRemote(c.param1);
        break;
      }
    }
  },

  mergeToLocal: function () {
    this.mFastSync = this.mClient.fastPossible;;
    if (this.mFastSync)
      this.fastMergeToLocal();
    else
      this.slowMergeToLocal();
    this.processList();
    this.mClient.finish();
  },

  mergeToRemote: function () {
    this.mFastSync = this.mClient.fastPossible;;
    if (this.mFastSync)
      this.fastMergeToRemote();
    else
      this.slowMergeToRemote();
    this.processList();
    this.mClient.finish();
  },

  sync: function () {
    this.mFastSync = this.mClient.fastPossible;;
    if (this.mFastSync) {
      this.fastMergeToLocal();
      this.fastMergeToRemote();
    } else {
      this.slowMergeToLocal();
      this.slowMergeToRemote();
    }
    this.processList();
    this.mClient.finish();
  },

  QueryInterface: function (iid) {
    if (!iid.equals(Components.interfaces.nsISyncSession) &&
        !iid.equals(Components.interfaces.nsISupports)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
  },

  mClient: null,
  mFastSync: false,

  mCommandList: null
}

var syncModule = {
  firstTime: true,

  registerSelf: function (compMgr, fileSpec, location, type) {
    if (this.firstTime) {
      this.firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    compMgr.registerFactoryLocation(this.syncSessionCID,
                                    "Sync Session JS Component",
                                    this.syncSessionProgID, 
                                    fileSpec,
                                    location,
                                    type);
  },

  getClassObject: function (compMgr, cid, iid) {
    if (cid.equals(this.syncSessionCID))
      return this.syncSessionFactory;

    if (!iid.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  syncSessionCID: Components.ID("{0536fac0-1358-4161-8405-35039eb985db}"),
  syncSessionProgID: "@mozilla.org/sync/session;1",
  syncSessionFactory: {
    createInstance: function (outer, iid) {
      if (outer != null)
        throw Components.results.NS_ERROR_NO_AGGREGATION;

      return (new nsSyncSession()).QueryInterface(iid);
    }
  },

  canUnload: function(compMgr) {
    return true;
  }
};

function NSGetModule(compMgr, fileSpec) {
  return syncModule;
}
